#ifndef GUARD_Dev_H
#define GUARD_Dev_H

#include "dg_node.h"
#include "debug.h"
#include "svr.h"
#include <string>
#include <unordered_set>
#include <vector>
#include <iostream>
#include <sys/time.h>

/*
	Dev class storing relavent information of detected (IoT and non-IoT) devices
*/

class Dev {
public:
 Dev(const std::string& m, bool t, size_t thre, time_t bt):MAC(m),isIoT(t),serverCtThre(thre),boot_time(bt),bs_abi_rate(-1.0),max_post_bs_abi_rate(-1.0),abi_def(true),bs_abi_ct(0),abi_turnoff_trial(0){}
 Dev(const std::string& m, bool t, size_t thre, const std::vector<const DG_Node*>& pMN, time_t bt):MAC(m),isIoT(t),serverCtThre(thre),boot_time(bt),bs_abi_rate(-1.0),max_post_bs_abi_rate(-1.0),abi_def(true),bs_abi_ct(0),abi_turnoff_trial(0){poMftrNodes=pMN;} 
 bool getIsIoT() const { return isIoT;}
 time_t getBootTime() const {return boot_time;}
 void setBootTime(time_t bt) {boot_time=bt;}
 const std::string& getMAC() const { return MAC;}
 int getServerCt() const { return servers.size();}
 int getAbnServerCt() const { return abn_servers.size();}
 void addServer(Svr* s){
	servers.insert(s); 
	if(s->abn){
		abn_servers.insert(s);
		if (isIoT && abn_servers.size()>serverCtThre){
			isIoT=false;
#if defined(DEBUG) || defined(LOG)
                 std::clog << "[LOG] re-classify IoT dev " << MAC << " as non-IoT" << std::endl;;		
#endif
		}
	} else {
		abi_servers.insert(s);
	}
 }
void removeServer(Svr* s){
	servers.erase(s);
        if(s->abn){
                abn_servers.erase(s);
                if (!isIoT && abn_servers.size()<=serverCtThre){
                        isIoT=true;
#if defined(DEBUG) || defined(LOG)
                 std::clog << "[LOG] re-classify non-IoT dev " << MAC << "as IoT" << std::endl;;  
#endif
                }
        } else {
		abi_servers.erase(s);
	}
}

 const std::vector<const DG_Node*>* getPoMftrNodes() const {return &poMftrNodes;}
 const std::unordered_set<Svr*>& getServers() const {return servers;}
 bool containServer(Svr* sr) const {return servers.find(sr)!=servers.end();}


//private:
 std::unordered_set<std:: string> bootstrapped_domains; // domain of bootstrapped:% legitemate IoT servers for this device
 std::unordered_set<Svr*> abi_servers; // set of abi servers this device talk to
 std::unordered_set<Svr*> wl_abi_servers; // set of whitelisted abi servers this device talk to
 std::unordered_set<Svr*> abn_servers; // set of abn servers this device talk to
 std::string MAC; // mac address of the device
 bool isIoT; // is this device an IoT 
 std::unordered_set<Svr*> servers; // set of servers this device talk to
 size_t serverCtThre; // the max number of distinct servers an IoT device should talk to
 time_t boot_time; // estimated boot/reboot time, initailized as capture timestamp of ths first outgoing packet from this device
 double bs_abi_rate; // average number of new ABI server dev visit per hour
 double max_post_bs_abi_rate; // average number of new ABI server dev visit per hour
 bool abi_def;
 size_t bs_abi_ct; // number of abi server this device talk to during bootstrapping
 std::vector<const DG_Node*> poMftrNodes; //storing potential manufacuters
 int abi_turnoff_trial;
};


#endif //GUARD_Dev_H
