#ifndef GUARD_IOTSTEED_H
#define GUARD_IOTSTEED_H

#include "dev.h"
#include "svr.h"
#include "dg_node.h"
#include <map>
#include <string>
#include <utility>
#include <sys/time.h>
#include <pcapplusplus/Packet.h>
#include <pcapplusplus/RawPacket.h>
#include <faup/faup.h>
#include <faup/options.h>
#include <fstream>


class IoTSTEED {
public:
 IoTSTEED(const char* mf_fn,const char* key,const char* pref, const char* exc_MACs, const char* wl_IPs, const char* router_MACs, size_t serverCtThre, const char* bsThre,const char* outfn_csv);
 void process_packet(pcpp::RawPacket&, faup_options_t*, int);


 std::string sha_prefix; //Shared: prefix of access network iotsteed is deployed
 size_t sha_serverCtThre; // the max number of distinct servers an IoT device should talk to 
 std::map<std::string, DG_Node> det_name2node; // IoT detection: the DG nodes storing prior knowledge of mftrs and mrlvs (with mftr/mrlvs as key)
 std::string det_apikey; // IoT detection: MAC address vendor lookup ibrary (https://macaddress.io/)
 std::map<std::string, Dev> det_mac2dev; // IoT detection: MAC address to IoT device mapping; to record every non-IoT and IoT devices we are tracking
 std::unordered_set<std::string> det_exception_MACs; // IoT detection: MACs for devices whose packets we ignore
 std::unordered_set<std::string> det_router_MACs; // IoT detection: MACs for LAN routers who we do not apply detection and learning
 std::unordered_set<std::string> sln_whitelisted_IPs; // Server learning: server IPs we white-list
 std::unordered_set<std::string> sln_whitelisted_s24; // Server learning: server /24 we white-list
 std::unordered_set<std::string> sln_whitelisted_s16; // Server learning: server /16 we white-list
 std::unordered_set<std::string> sln_whitelisted_s8; // Server learning: server /8 we white-list
 std::map<std::string, Svr> sln_sname2svr; // Server Learning: the master copy of Svr objects for all server names seen so far (with sname as key)
 std::map<std::string, Svr*> sln_cname2svr; // Server Learning: server CNAME seen so far, mapped to Svr object
 std::map<std::string, Svr*> sln_IP2svr; // Server Learning: IP seen so far, mapped to Svr object
 time_t sln_bsThre_abn; //Server Learning: bootstrapping period for ABN servers
 time_t sln_bsThre_abi; //Server Learning: bootstrapping period for ABI servers
 std::ofstream def_of; // Defense: output packet_num, timstamp_sec+ 7 tuples for dropped packets

private:
 void detect_from_MAC(std::string, time_t);
 void apply_detection(time_t, std::string&, std::string&);
 std::pair<Dev*,std::string*> apply_learning(int, int, int, time_t, std::string&, std::string&, std::string&, std::string&, pcpp::RawPacket&, pcpp::Packet&, faup_options_t*);
 void apply_defense(Dev*, std::string*, int, int, int, time_t, std::string&, std::string&, int, std::string&, std::string&);
};


#endif //GUARD_IOTSTEED_H

