#include "svr.h"
#include "sln.h"
#include <map>
#include <string>
#include <iostream>
#include <algorithm>
#include <pcapplusplus/Packet.h>
#include <pcapplusplus/DnsLayer.h>
#include <pcapplusplus/DnsResource.h>
#include <pcapplusplus/DnsResourceData.h>
#include <faup/faup.h>
#include <faup/decode.h>
#include <faup/options.h>
#include "debug.h"


using namespace std;
using namespace pcpp;

/*
	return if S1 is a substring of S2, regardless of case
*/
bool is_substr_ci(string S1, string S2){
	transform(S1.begin(), S1.end(), S1.begin(), ::tolower); 
	transform(S2.begin(), S2.end(), S2.begin(), ::tolower); 
	return S2.find(S1) != std::string::npos;
}

/*
        replace digits in server name as matching any digit
*/
void replace_digit_with_asterisk(string& sname){
        for (string::size_type i = 0; i <= sname.length( ); i++)
        {
                if (sname[i] >= '0' && sname[i] <= '9')
                {
                        sname.at(i)='*';
                }
        }
}

/*
	Knowing lan_iot DNS query/talk to wan_svr at pkt_timestamp for the FIRST time
        Apply two rounds of server learning.
*/

void two_round_server_learning(Dev* lan_iot, Svr* wan_svr, time_t pkt_timestamp, time_t sln_bsThre_abn, time_t sln_bsThre_abi, faup_options_t* faup_opts)
{

	/* 
		Sanity check if its the first time lan_iot DNS query/talk to wan_svr 
	*/
	if((wan_svr->dev2svr_cls).find(lan_iot)!=(wan_svr->dev2svr_cls).end())
	{ 
#ifdef DEBUG 
		clog << "[Warn] two_round_server_learning(): server " << wan_svr->sname << " already leanrt for device " << lan_iot->getMAC() << endl;
#endif
	}

	//record first time lan_iot  DNS query/talk to wan_svr
        (wan_svr->dev2first_access_time).insert(pair<Dev*, time_t>(lan_iot, pkt_timestamp));
	


        /*
                first extract domain from wan_svr->sname; (for abi-servers, sname=sname_domain=server IP)
        */
        faup_handler_t *fh;
        if (!faup_opts) { cerr << "[Error] two_round_server_learning(): cannot allocate faup options" << endl;}
        fh = faup_init(faup_opts);
        string sname = wan_svr->sname;
        faup_decode(fh, sname.c_str(), sname.size() );
        string sname_domain = sname.substr(faup_get_domain_without_tld_pos(fh),faup_get_domain_without_tld_size(fh));
        transform(sname_domain.begin(), sname_domain.end(), sname_domain.begin(), ::tolower); //sname to lower
        sname_domain.erase(remove_if(sname_domain.begin(), sname_domain.end(), [](char c) { return ! (isalnum(c) || isspace(c)) ; }), sname_domain.end()); // removing any special character from sname: leaving only alpha, num, space. Rationale is when we store mftr internally, we also remove all special character
        faup_terminate(fh);
	time_t sln_bsThre;

	if(wan_svr->abn){
		sln_bsThre=sln_bsThre_abn;
#ifdef DEBUG 
                clog << "###Use ABN server bootstrapping threshold: " << sln_bsThre << " secs" << endl;
#endif
	} else {
		sln_bsThre=sln_bsThre_abi;
#ifdef DEBUG 
                clog << "###Use ABI server bootstrapping threshold: " << sln_bsThre << " secs" << endl;
#endif
	}

	if(pkt_timestamp <= lan_iot->getBootTime()+sln_bsThre) // lan_iot query/visit wan_svr shortly after (re-)boot. First-round learning: wan_svr is legit.
	{	
#ifdef DEBUG 
		clog << "###Entered First-round Learning: " << sname << " is IoT server for device " << lan_iot->getMAC() << " boot at " << lan_iot->getBootTime()<< endl;
#endif
		(wan_svr->dev2svr_cls).insert(pair<Dev*, bool>(lan_iot, true));
		(wan_svr->dev2rounds).insert(pair<Dev*, double>(lan_iot, 1.0));
		lan_iot->bootstrapped_domains.insert(sname_domain);

		return;

	} else { // lan_iot query/visit wan_svr long after (re-)boot. Second-round learning: examine wan_svr's sname.

#ifdef DEBUG 
                clog << "###Entered Second-round Learning: testing domain " << sname_domain << " of server " << sname << " for device " << lan_iot->getMAC() << " boot at " << lan_iot->getBootTime()<< endl;
#endif

		if(!wan_svr->abn)
		{
			if(lan_iot->bs_abi_rate==-1) // if wan_svr is the first new, non-whitelisted ABI server lan_iot talk to after bootstrapping
			{ 
				lan_iot->bs_abi_ct=lan_iot->abi_servers.size()-lan_iot->wl_abi_servers.size()-1; //only count non-whitelisted ABI servers in bs_abi_ct and bs_abi_rate
				lan_iot->bs_abi_rate=lan_iot->bs_abi_ct*1.0/sln_bsThre_abi;
#ifdef DEBUG 
		                clog << "###First new ABI server after bootstrapping: avg bootstrapping abi server rate=" << lan_iot->bs_abi_rate << endl;
#endif
			}

			time_t elasped_post_bs=pkt_timestamp-lan_iot->getBootTime()-sln_bsThre_abi;
			double post_bs_abi_rate=(lan_iot->abi_servers.size()-lan_iot->wl_abi_servers.size()-lan_iot->bs_abi_ct)*1.0/elasped_post_bs; //only count non-whitelisted ABI servers in post_bs_abi_rate
                        lan_iot->max_post_bs_abi_rate=max(lan_iot->max_post_bs_abi_rate,post_bs_abi_rate);
#ifdef DEBUG                            
                        clog << "###Current avg post-bootstrapping abi server rate=" << post_bs_abi_rate << endl;
                        clog << "###Current max avg post-bootstrapping abi server rate=" << lan_iot->max_post_bs_abi_rate << endl;
#endif
			if(lan_iot->abi_def==true && lan_iot->bs_abi_ct>=turn_off_min_abi_ct && post_bs_abi_rate>turn_off_ratio*lan_iot->bs_abi_rate){
				lan_iot->abi_turnoff_trial+=1;
				if(lan_iot->abi_turnoff_trial>=turn_off_delay){
					lan_iot->abi_def=false;
#if defined(DEBUG) || defined(LOG)
	                       	 	clog << "[LOG] " << pkt_timestamp <<  ": turn off ABI Defense for IoT device " << lan_iot->getMAC() << endl;
#endif
				}

			}


			if(!lan_iot->abi_def) { // whitelist all new ABI servers after turn off abi_def as benign
#ifdef DEBUG 
                                clog << "###Hit: server " << sname << " is whitelisted as IoT because ABI defense is off for device " << lan_iot->MAC << endl;
#endif
                                (wan_svr->dev2svr_cls).insert(pair<Dev*, bool>(lan_iot, true));
                                (wan_svr->dev2rounds).insert(pair<Dev*, double>(lan_iot, 2.0));
				return;
			}
		}
 
                /*
                        Is wan_svr a manufacuter server for lan_iot?
                */
		for(vector<const DG_Node*>::const_iterator it=lan_iot->getPoMftrNodes()->begin(); it != lan_iot->getPoMftrNodes()->end(); it++)
		{
			if(is_substr_ci((*it)->getName(),sname_domain)) // if any potential mftr is a substr of sname's domain?
			{
#ifdef DEBUG 
                		clog << "###Hit: server " << sname << " is an IoT mftr server" << endl;
#endif
				(wan_svr->dev2svr_cls).insert(pair<Dev*, bool>(lan_iot, true));
		                (wan_svr->dev2rounds).insert(pair<Dev*, double>(lan_iot, 2.1));
				return;
			}

		}

		/*
			is wan_svr a known platform server?
		*/

		if(platform_domains.find(sname_domain)!=platform_domains.end())
		{
#ifdef DEBUG 
                	clog << "###Hit: server " << sname << " is an IoT platform server" << endl;
#endif
                       (wan_svr->dev2svr_cls).insert(pair<Dev*, bool>(lan_iot, true));
                       (wan_svr->dev2rounds).insert(pair<Dev*, double>(lan_iot, 2.2));
		       return;
		}

                /*
                        is wan_svr a known third-party server?
                */

                if(lan_iot->bootstrapped_domains.find(sname_domain)!=lan_iot->bootstrapped_domains.end() || sname_domain.compare("ntp")==0 || sname_domain.compare("nist")==0 || sname.find("time") !=string::npos || sname.find("ntp") !=string::npos || sname.find("clock") !=string::npos)
                {
#ifdef DEBUG 
                        clog << "###Hit: server " << sname << " is an IoT 3rd-party server" << endl;
#endif
                       (wan_svr->dev2svr_cls).insert(pair<Dev*, bool>(lan_iot, true));
                       (wan_svr->dev2rounds).insert(pair<Dev*, double>(lan_iot, 2.3));
                       return;
                }

		
		// wan_svr is not a legitemate IoT server
		(wan_svr->dev2svr_cls).insert(pair<Dev*, bool>(lan_iot, false));
                (wan_svr->dev2rounds).insert(pair<Dev*, double>(lan_iot, 3.0));
                cout << "[Warn] " << pkt_timestamp << ": IoT dev " << lan_iot->MAC << " visit suspecious svr " << wan_svr->sname << endl;
	 	//cout << "      This device is likely comprpomised. Please reboot or factory reset this device, modify device login password and update device firmware." << endl;

#ifdef DEBUG 
                        clog << "###Miss: server " << sname << " is a non-IoT server" << endl;
#endif


	}

}


/* 
	        Handle DNS Query from LAN IoT:
                (1) Create a Svr obj for every abn-servers (by identifying every queried, distinct sname)
                (2) Record a list of abn-servers Svr objs that each IoT query/will talk to
                (3) 2-round learning of abn-servers for each LAN IoT querying them

*/

void track_DNS_query(Packet& parsedPacket, Dev& sdev_iot, map<string, Svr>& sln_sname2svr, map<string,Svr*>& sln_cname2svr, time_t pkt_timestamp, time_t sln_bsThre_abn, time_t sln_bsThre_abi, faup_options_t* faup_opts)
{
	DnsLayer*  dnsLayer = parsedPacket.getLayerOfType<DnsLayer>();

	if(dnsLayer==NULL)
        {
#ifdef DEBUG 
        	clog << "##Skipped malformed DNS queries without DNS header" << endl;
#endif
                return;
        }

	size_t q_ct=dnsLayer->getQueryCount();

	DnsQuery * q_cur=NULL;

	for(size_t i=0; i<q_ct; i++)
	{
		if(q_cur==NULL)
		{
			q_cur=dnsLayer->getFirstQuery();
		}
		else
		{
			DnsQuery* q_next=dnsLayer->getNextQuery(q_cur);
			q_cur=q_next;
		}
		string qname=q_cur->getName();
		replace_digit_with_asterisk(qname);
		DnsType qtype=q_cur->getDnsType();
		Svr* cur_svr=NULL;

#ifdef DEBUG
                clog << "##Quesion " << i+1 << ": Q_name=" << qname << "; Q_type=" << qtype  << endl;
#endif
		if(qtype!=DNS_TYPE_A && qtype!=DNS_TYPE_CNAME && qtype!=DNS_TYPE_AAAA ){
#ifdef DEBUG 
	                clog << "###Skipped non-A-CNAME-AAAA DNS queries" << endl;
#endif
		} else if(sln_sname2svr.find(qname)!=sln_sname2svr.end()){ //qname is a seen sname, retrive corresponding Svr obj in cur_svr

#ifdef DEBUG
			clog << "###QName is an exisiting SName" << endl;
#endif
			cur_svr=&(sln_sname2svr.find(qname)->second);

		} else if (sln_cname2svr.find(qname)!=sln_cname2svr.end()){ // qname is a seen cname, retrive corresponding Svr obj in cur_svr
#ifdef DEBUG
                        clog << "###QName is an exisiting CName"  << endl;
#endif		
			cur_svr=sln_cname2svr.find(qname)->second;			

		} else { // qname is a new sname; create a new Svr obj; store it in sln_sname2svr; point to it with cur_svr
#ifdef DEBUG
                        clog << "###QName is a new SName" << endl;
#endif

		        faup_handler_t *fh;
        		if (!faup_opts) { cerr << "[Error] track_DNS_query(): cannot allocate faup options" << endl;}
        		fh = faup_init(faup_opts);
        		faup_decode(fh, qname.c_str(), qname.size() );
			if((size_t) faup_get_tld_pos(fh)<=qname.length()){
				string qname_tld = qname.substr(faup_get_tld_pos(fh),faup_get_tld_size(fh));
				if(invalid_TLDs.find(qname_tld)!=invalid_TLDs.end())
				{
#ifdef DEBUG
                        		clog << "###Ignoreing QName with invalid TLD" << qname_tld << endl;
#endif
					return;
				}
			}

			Svr new_svr(qname,true);
			sln_sname2svr.insert(pair<string,Svr>(qname,new_svr));
			cur_svr=&(sln_sname2svr.find(qname)->second);
		}

		//if its the first time sdev_iot DNS query cur_svr, record that and apply two_round server learning to cur_svr
		if(cur_svr!=NULL && !sdev_iot.containServer(cur_svr))
		{
#ifdef DEBUG
                        clog << "###First time Svr " << cur_svr->sname << " get accessed by device " << sdev_iot.getMAC() << endl;
#endif

                        sdev_iot.addServer(cur_svr);
			two_round_server_learning(&sdev_iot, cur_svr, pkt_timestamp, sln_bsThre_abn, sln_bsThre_abi, faup_opts);
		}
	}

} 


/* 
		Track IoT DNS resolution: extract IPs and CNames for servers names IoT device query
                (Specifically, by parsing DNS replies, fill up IP and CName for each Svr objs)

*/
void track_DNS_reply(Packet& parsedPacket, Dev& ddev_iot, map<string, Svr>& sln_sname2svr, map<string, Svr*>& sln_cname2svr, map<string, Svr*>& sln_IP2svr,time_t sln_bsThre_abn, time_t sln_bsThre_abi, faup_options_t* faup_opts,time_t pkt_timestamp)
{

        DnsLayer*  dnsLayer = parsedPacket.getLayerOfType<DnsLayer>();

        if(dnsLayer==NULL)
        {
#ifdef DEBUG 
                clog << "##Skipped malformed DNS replies without DNS header" << endl;
#endif
                return;
       }


      size_t a_ct=dnsLayer->getAnswerCount();

      DnsResource* a_cur=NULL;

      for(size_t i=0; i<a_ct; i++)
      {
              if(a_cur==NULL)
              {
                      a_cur=dnsLayer->getFirstAnswer();
              }
              else
              {
                      DnsResource* a_next=dnsLayer->getNextAnswer(a_cur);
                      a_cur=a_next;
              }

      		string aname=a_cur->getName();
		replace_digit_with_asterisk(aname);
      		DnsType atype=a_cur->getDnsType();
      		DnsResourceDataPtr a_cur_dp=a_cur->getData();

      		if(a_cur_dp.isTypeOf<IPv4DnsResourceData>())
      		{
      		 	IPv4DnsResourceData* a_cur_data=a_cur_dp.castAs<IPv4DnsResourceData>();
			string new_IP(a_cur_data->toString());
#ifdef DEBUG
	         	clog << "###Answer " << i+1 << ": A_owner=" << aname << "; A_type=" << atype << endl;
#endif
			if(atype!=DNS_TYPE_A) 
			{	//shoudn't be here
				cerr << "[Error] non-type-A DNS answer with IPv4 rdata" << endl;

			} else if (sln_IP2svr.find(new_IP)==sln_IP2svr.end()) {//new IP; update corresponding Svr
#ifdef DEBUG
                        	clog  << "####New IP " << new_IP << ": Update corresponding Svr obj" << endl;
#endif
				Svr* svr_ptr=NULL; //pointer to svr corresponding to aname

				if(sln_sname2svr.find(aname)!=sln_sname2svr.end()){ // Aname is a SName
#ifdef DEBUG
                        		clog << "####A_owner " << aname << " match a knwon SName" << endl;
#endif
					svr_ptr=&(sln_sname2svr.find(aname)->second);

				} else if (sln_cname2svr.find(aname)!=sln_cname2svr.end()){ //Aname is a Cname
#ifdef DEBUG
                                        clog << "####A_owner " << aname << " match a knwon CName" << endl;
#endif
					svr_ptr=sln_cname2svr.find(aname)->second;
				}

				if(svr_ptr)
				{
                                       svr_ptr->addIP(new_IP);
                                       sln_IP2svr.insert(pair<string,Svr*>(new_IP, svr_ptr));

                                       if(sln_sname2svr.find(new_IP)!=sln_sname2svr.end()) //if svr_ptr was fasely considered abi-server, likely due to out-of-order DNS traffic
				       {
						Svr& abi_svr=sln_sname2svr.find(new_IP)->second;
						for(map<Dev*, time_t>::iterator it=abi_svr.dev2first_access_time.begin(); it!=abi_svr.dev2first_access_time.end(); it++){  //modify all dev talk to abi_svr as if they talk to svr_ptr in the first place
							Dev* cdev=it->first;
                                                        time_t abi_access_time=it->second;

							//remove record that cdev visit abi_svr
                                                        cdev->removeServer(&abi_svr);
							if((abi_svr.dev2rounds.find(cdev))->second==0) //if abi_svr is a whitelisted abi server for cdev
							{	
								cdev->wl_abi_servers.erase(&abi_svr);	
							} else if (abi_access_time<=cdev->boot_time+sln_bsThre_abi) //if abi_svr is non-whitelisted and visited during bs, update bs_abi_rate and bs_abi_ct
							{
								cdev->bs_abi_ct-=1;
								cdev->bs_abi_rate=cdev->bs_abi_ct*1.0/sln_bsThre_abi;
								if(cdev->bs_abi_ct<turn_off_min_abi_ct)
								{
									if(cdev->abi_turnoff_trial>0){
#if defined(DEBUG) || defined(LOG)
					                                        clog << "[LOG] " << pkt_timestamp <<  ": one abi_svr during bs get re-cls as abn; reset abi_turnofff_trial for IoT device " << cdev->getMAC() << endl;
#endif
										cdev->abi_turnoff_trial=0;
									}
									if(cdev->abi_def==false){
#if defined(DEBUG) || defined(LOG)
                                                                                clog << "[LOG] " << pkt_timestamp <<  ": one abi_svr during bs get re-cls as abn; turn back on ABI defense for IoT device " << cdev->getMAC() << endl;
#endif
										cdev->abi_def=true;
									}
								}

							} else { //if abi_svr is non-whitelisted and visited during bs, mark current max_post_bs_abi_rate may be wrong
								cdev->max_post_bs_abi_rate=9999;
							}

							// make as if cdev talk to svr_ptr at first place (at abi_access_time)
                                                        cdev->addServer(svr_ptr);
							two_round_server_learning(cdev, svr_ptr, abi_access_time, sln_bsThre_abn, sln_bsThre_abi, faup_opts);
						}
						sln_sname2svr.erase(new_IP); // drop this flase abi_svr
				       }


				} else {

#ifdef DEBUG
					cerr << "[Warn] no know SNAME or CNAME matching A_owner " << aname << endl;
#endif
				}


			} else {// existing IP; do nothing
#ifdef DEBUG
                                clog  << "####Existing IP " << new_IP << ": do nothing" << endl;
#endif
			}

		} else if (a_cur_dp.isTypeOf<StringDnsResourceData>()){

		 	StringDnsResourceData* a_cur_data=a_cur_dp.castAs<StringDnsResourceData>();
			string new_cname(a_cur_data->toString());
			replace_digit_with_asterisk(new_cname);

#ifdef DEBUG
                 	clog << "###Answer " << i+1 << ": A_owner=" << aname << "; A_type=" << atype << endl;
#endif

                        if(atype!=DNS_TYPE_CNAME){

#ifdef DEBUG
                                clog << "Ignore non-CNAME DNS answer with string rdata" << endl;
#endif

                        } else if (sln_cname2svr.find(new_cname)==sln_cname2svr.end()) {//new cname; update corresponding Svr

#ifdef DEBUG
                                clog  << "####New Cname " << new_cname << ": Update corresponding Svr obj" << endl;
#endif

                                Svr* svr_ptr=NULL; //pointer to svr corresponding to aname

                                if(sln_sname2svr.find(aname)!=sln_sname2svr.end()){ // Aname is a SName
#ifdef DEBUG
                                        clog << "####A_owner " << aname << " match a knwon SName" << endl;
#endif
                                        svr_ptr=&(sln_sname2svr.find(aname)->second);

                                } else if (sln_cname2svr.find(aname)!=sln_cname2svr.end()){ //Aname is a Cname
#ifdef DEBUG
                                        clog << "####A_owner " << aname << " match a knwon CName" << endl;
#endif
                                        svr_ptr=sln_cname2svr.find(aname)->second;
                                }

                                if(svr_ptr)
                                {
                                       svr_ptr->addCname(new_cname);
                                       sln_cname2svr.insert(pair<string,Svr*>(new_cname, svr_ptr));

                                } else {
#ifdef DEBUG
                                        cerr << "[Warn] no know SNAME or CNAME matching A_owner " << aname << endl;
#endif
                                }
	

			} else {//exisiting cname; do nothing

#ifdef DEBUG
                                clog  << "####Existing Cname " << new_cname << ": do nothing" << endl;
#endif
			}


		} else {

#ifdef DEBUG
                       clog << "###Answer " << i+1 << ": A_owner=" << aname << "; A_type=" << atype << endl;
		       clog << "####Ignore non-A-CNAME answer";
#endif

		}
      }
}

/*
                For all LAN-IoT-WAN-Server traffic:         (1) Crate a Svr obj for every abi-servers
                (including trafic to WAN DNS server IP)     (2) Record a list of abi-servers Svr objts that each IoT talk to
                                                            (3) 2-round learning of abi-servers that LAN IoT talk to
replace_digit_with_asterisk
*/
void handle_lan_iot_wan_server_traffic(Dev* lan_iot, string* wan_svr_IP, unordered_set<string>& sln_whitelisted_IPs, unordered_set<string>& sln_whitelisted_s24, unordered_set<string>& sln_whitelisted_s16, unordered_set<string>& sln_whitelisted_s8, map<string, Svr>& sln_sname2svr, map<string,Svr*>& sln_IP2svr, time_t pkt_timestamp, time_t sln_bsThre_abn, time_t sln_bsThre_abi, faup_options_t* faup_opts)
{
        if(sln_IP2svr.find(*wan_svr_IP)==sln_IP2svr.end()) // wan_svr is an abi-server
        {
                if(sln_sname2svr.find(*wan_svr_IP)==sln_sname2svr.end()) // new abi-server
		{
#ifdef DEBUG 
      	             clog << "##New abi-server " << *wan_svr_IP << endl;
#endif
                     Svr new_svr(*wan_svr_IP,false); //ceate a Svr obj for wan_svr with IP as identifier
                     sln_sname2svr.insert(pair<string,Svr>(*wan_svr_IP,new_svr));

		} else {
#ifdef DEBUG 
           	     clog << "##Exisiting abi-server " << *wan_svr_IP  << endl;
#endif
		}

                Svr* wan_svr=&(sln_sname2svr.find(*wan_svr_IP)->second);

        	if(!lan_iot->containServer(wan_svr)){ // if first time lan_iot talk to wan_svr, record lan_iot talk to wan_svr and apply learning 
#ifdef DEBUG 
	                clog << "##First time abi-server " << *wan_svr_IP << " get visited by device " << lan_iot->getMAC() << endl;
#endif

                	lan_iot->addServer(wan_svr);
		
			string IP=string(*wan_svr_IP);
		        string S24=IP.substr(0,IP.find_last_of("."));
        		string S16=S24.substr(0,S24.find_last_of("."));
        		string S8=S16.substr(0,S16.find_last_of("."));


                	if(sln_whitelisted_IPs.find(IP) != sln_whitelisted_IPs.end() || sln_whitelisted_s24.find(S24) != sln_whitelisted_s24.end() ||  sln_whitelisted_s16.find(S16) != sln_whitelisted_s16.end() ||  sln_whitelisted_s8.find(S8) != sln_whitelisted_s8.end() ) //if white-listed server IPs
                	{
#ifdef DEBUG 
          		      clog << "###Abi-server " << *wan_svr_IP << " is whitelisted as an IoT server" << endl;
#endif    
                        	(wan_svr->dev2svr_cls).insert(pair<Dev*, bool>(lan_iot, true));
				(wan_svr->dev2first_access_time).insert(pair<Dev*, time_t>(lan_iot, pkt_timestamp));
				(wan_svr->dev2rounds).insert(pair<Dev*, double>(lan_iot, 0.0));
				lan_iot->wl_abi_servers.insert(wan_svr);
                	} else { //if not, send abi-server to two-round learning (abi-server can only pass first round  because its sname is IP)
                        	two_round_server_learning(lan_iot, wan_svr, pkt_timestamp, sln_bsThre_abn, sln_bsThre_abi, faup_opts);
                	}
        	}


        } else { // wan_svr is an abn-server: need to handle cover case that lan_iot first visit a abn-server by IP and needs server leanring
#ifdef DEBUG 
                clog << "##WAN server is an exisiting abn-server" << endl;
#endif
		Svr* wan_svr=sln_IP2svr.find(*wan_svr_IP)->second;
		if(!lan_iot->containServer(wan_svr)){ // if first time  lan_iot talk to wan_svr, record lan_iot talk to wan_svr and apply learning 
#ifdef DEBUG 
                        clog << "##First time abn-server " << wan_svr->sname << " get visited by device " << lan_iot->getMAC() << endl;
#endif
                        lan_iot->addServer(wan_svr);
			two_round_server_learning(lan_iot, wan_svr, pkt_timestamp, sln_bsThre_abn, sln_bsThre_abi, faup_opts);
		}

        }

}
