#ifndef GUARD_Sln
#define GUARD_Sln

#include <map>
#include <string>
#include "svr.h"
#include "dev.h"
#include <sys/time.h>
#include <unordered_set>
#include <pcapplusplus/Packet.h>
#include <faup/faup.h>
#include <faup/decode.h>

/*
	constant values used in server learning
*/

const std::unordered_set<std::string> platform_domains ({"tuyacn", "tuyaeu", "tuyaus", "evrythng", "pubnub", "pndsn", "xively", "azuredevices"});  // list of known domains for IoT platform servers. s/azure-devices/azuredevices/ becuase we remove all non-letter-digit character from sname_domain in sln.cc
const std::unordered_set<std::string> invalid_TLDs ({"test","example","invalid","localhost","local","localdomain","domain","lan","home","host","corp"}); //list of invalid TLD from https://tools.ietf.org/id/draft-chapin-rfc2606bis-00.html

const int turn_off_delay=1; //turn off ABI defense after condition met turn_off_delay times
const size_t turn_off_min_abi_ct=3; //minimum abi server count during bootstrapping required for turnning off ABI defense
const double turn_off_ratio=0.5; ///when post_bs_abi_server_rate > turn_off_ratio* bs_abi_server_rate, we turn off ABI defense


/*
	functions related to server learning
*/

/* apply two-round server learning*/
void two_round_server_learning(Dev*, Svr*, time_t, time_t, time_t, faup_options_t*);

/*Track server name IoT devices talk to from DNS queries*/
void track_DNS_query(pcpp::Packet&, Dev& , std::map<std::string, Svr>&, std::map<std::string, Svr*>&, time_t, time_t, time_t, faup_options_t*);

/*Track CNAME and IPs for server names IoT devices talk to from DNS replies*/
void track_DNS_reply(pcpp::Packet&, Dev& , std::map<std::string, Svr>& , std::map<std::string, Svr*>&, std::map<std::string, Svr*>&, time_t,time_t, faup_options_t*,time_t);

/*Handle traffic between LAN-IoT and WAN-Server*/
void handle_lan_iot_wan_server_traffic(Dev*, std::string*, std::unordered_set<std::string>&, std::unordered_set<std::string>&, std::unordered_set<std::string>&, std::unordered_set<std::string>&, std::map<std::string, Svr>&, std::map<std::string,Svr*>& , time_t , time_t , time_t, faup_options_t* );

#endif //GUARD_Sln

